@@ -18,7 +18,7 @@
 */
 
 /*
- * $Id: ccid.c 6274 2012-04-13 13:45:16Z rousseau $
+ * $Id: ccid.c 6650 2013-06-10 08:43:24Z rousseau $
  */
 
 #include <stdio.h>
@@ -33,6 +33,7 @@
 #include "defs.h"
 #include "ccid_ifdhandler.h"
 #include "commands.h"
+#include "utils.h"
 
 #ifdef __APPLE__
 #include <CoreFoundation/CoreFoundation.h>
@@ -64,6 +65,19 @@ int ccid_open_hack_pre(unsigned int reader_index)
 			(void)sleep(1);
 			ccid_descriptor->readTimeout = 60*1000; /* 60 seconds */
 			break;
+
+		case GEMPCTWIN:
+		case GEMPCKEY:
+		case DELLSCRK:
+			/* Only the chipset with firmware version 2.00 is "bogus"
+			 * The reader may send packets of 0 bytes when the reader is
+			 * connected to a USB 3 port */
+			if (0x0200 == ccid_descriptor->IFD_bcdDevice)
+			{
+				ccid_descriptor->zlp = TRUE;
+				DEBUG_INFO("ZLP fixup");
+			}
+			break;
 	}
 
 	/* CCID */
